/*
 * Copyright (c) 1982, 1986, 1991 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	@(#)ax_monitor.c	0.0 (Skinnarila) 22/5/94
 */

#include "param.h"
#include "systm.h"
#include "kernel.h"
#include "conf.h"
#include "malloc.h"
#include "ioctl.h"
#include "fcntl.h"
#include "stat.h"
#include "proc.h"
#include "user.h"
#include "uio.h"
#include "vnode.h"
#include "mbuf.h"
#include "socket.h"
#include "socketvar.h"
#include "protosw.h"
#include "domain.h"

#include "net/if.h"
#include "net/route.h"
#include "net/af.h"
#include "net/netisr.h"

#include "ax25/ax.h"
#include "ax25/ax25.h"
#include "ax25/ax_pcb.h"
#include "ax25/ax_var.h"

int ax_monitored;

static struct {
	int fifor;
	int fifow;
	u_char fifo[1024];

	struct selrecord sel;

	char asleep;
} axmon;

/* ARGSUSED */
axmonopen(dev, flag, mode, p)
	dev_t dev;
	int flag, mode;
	struct proc *p;
{
	++ax_monitored;

	return (0);
}

/* ARGSUSED */
axmonclose(dev, flag, mode, p)
	dev_t dev;
	int flag, mode;
	struct proc *p;
{

	selclear(&axmon.sel);
	ax_monitored = 0;

	return (0);
}

/* ARGSUSED */
axmonread(dev, uio, flag)
	dev_t dev;
	struct uio *uio;
{
	int s, error = 0;

	s = splimp();
	while (axmon.fifor == axmon.fifow){
		if (flag & IO_NDELAY){
			error = EWOULDBLOCK;
			break;
		}
		axmon.asleep = 1;
		error = tsleep((caddr_t) &axmon.fifow, PWAIT | PCATCH, "axmon", 0);
		if (error)
			break;
	}
	splx(s);
	if (error)
		return error;

	while (uio->uio_resid > 0){
		int cc = axmon.fifow - axmon.fifor;
		if (cc < 0)
			cc = sizeof(axmon.fifo) - axmon.fifor;
		if (cc > uio->uio_resid)
			cc = uio->uio_resid;
		if (cc <= 0)
			break;
		error = uiomove((caddr_t) &axmon.fifo[axmon.fifor], cc, uio);
		if (error)
			break;
		if (axmon.fifor + cc >= sizeof(axmon.fifo))
			axmon.fifor = 0;
		else
			axmon.fifor += cc;
	}
	return (error);
}

/* ARGSUSED */
axmonioctl(dev, cmd, data, flag)
	dev_t dev;
	int *data;
{
	int cc;

	switch(cmd){
	case FIONREAD:
		cc = axmon.fifow - axmon.fifor;
		if (cc < 0)
			cc += sizeof(axmon.fifo);
		*data = cc;
		return (0);

	case FIONBIO:
		return (0); /* accept the notification from upper layers */

	}
	return (ENOTTY);
}

/* ARGSUSED */
axmonselect(dev, which, p)
	dev_t dev;
	struct proc *p;
{
	int cc;

	if (which != FREAD)
		return (0); /* never ready for nothing but reading */

	cc = axmon.fifow - axmon.fifor;
	if (cc)
		return (1);

	selrecord(&axmon.sel, p);
	return (0);
}

static
addc(c)
	char c;
{
	int i = axmon.fifow;

	axmon.fifo[i++] = c;
	if (i >= sizeof(axmon.fifo))
		i = 0;
	if (i != axmon.fifor)
		axmon.fifow = i;
}
static
adds(s)
	char *s;
{
	while (*s)
		addc(*s++);
}
static
addd(d)
	unsigned d;
{
	char buf[16], *cp = buf;

	do {
		*cp++ = d % 10;
	} while (d /= 10);
	do {
		addc(*--cp + '0');
	} while (cp > buf);
}
static
addx(x)
	u_char x;
{
	static char hexes[] = "0123456789ABCDEF";

	addc(hexes[x >> 4]);
	addc(hexes[x & 15]);
}
static
stamp()
{
	int sec = time.tv_sec % 100;

	addc((sec / 10) + '0');
	addc((sec % 10) + '0');
	addc(' ');
}

ax_monitor(title, m)
	char *title;
	struct mbuf *m;
{
	struct axhdr *axh = mtod(m, struct axhdr *);
	u_char cmd = (axh->ah_dst.ax_ssid & SSID_C);
	u_char ctrl = axh->ah_ctrl;

	stamp();

	adds(ax_sprintf(&axh->ah_src));
	addc(' ');
	adds(ax_sprintf(&axh->ah_dst));
	addc('\t');

	switch (ctrl & 3) {

	default:
		adds("INFO ");
		addc((ctrl & PF) ? (cmd ? 'P' : 'F') : (cmd ? 'C' : 'R'));
		adds(" r=");
		addd(ctrl >> 5);
		adds(" s=");
		addd((ctrl >> 1) & 0x7);
		adds(" p=");
		addx(axh->ah_pid);
		addc(' ');
		addd(m->m_pkthdr.len - AXMINLEN - 1);
		break;

	case (SABM & 3):
		switch (ctrl & ~PF) {
		default:      adds("U??? "); break;
		case UI:      adds("UI   "); break;
		case SABM:    adds("SABM "); break;
		case UA:      adds("UA   "); break;
		case DISC:    adds("DISC "); break;
		case FRMR:    adds("FRMR "); break;
		case DM:      adds("DM   "); break;
		}
		addc((ctrl & PF) ? (cmd ? 'P' : 'F') : (cmd ? 'C' : 'R'));

		if ((ctrl & ~PF) == UI) {
			adds(" p=");
			addx(axh->ah_pid);
			addc(' ');
			addd(m->m_pkthdr.len - AXMINLEN - 1);
		}
		break;

	case (RR & 3):
		switch (ctrl & ~(NR|PF)) {
		default:      adds("S??? "); break;
		case RR:      adds("RR   "); break;
		case RNR:     adds("RNR  "); break;
		case REJ:     adds("REJ  "); break;
		}
		addc((ctrl & PF) ? (cmd ? 'P' : 'F') : (cmd ? 'C' : 'R'));
		adds(" r=");
		addd(ctrl >> 5);
		break;
	}
	addc('\n');

	if (axmon.asleep){
		axmon.asleep = 0;
		wakeup((caddr_t) &axmon.fifow);
	}
	selwakeup(&axmon.sel);
}

ax_event(axp, what)
	struct axpcb *axp;
	char *what;
{
	int dotimers = 1;

	stamp();

	adds(ax_sprintf(&axp->axp_laddr));
	addc(' ');
	adds(ax_sprintf(&axp->axp_faddr));
	addc('\t');

	if (what)
		adds(what);

	if (dotimers) {
		adds(" t=");
		addd(axp->axp_T1);
		addc(',');
		addd(axp->axp_T2);
		addc(',');
		addd(axp->axp_T3);
		adds(" n=");
		addd(axp->axp_N2);
		adds(" rtt=");
		addd(axp->axp_lastrtt);
		addc(',');
		addd(axp->axp_srtt);
	}

	addc('\n');

	if (axmon.asleep){
		axmon.asleep = 0;
		wakeup((caddr_t) &axmon.fifow);
	}
	selwakeup(&axmon.sel);
}

