
#include "standard.h"
#include "fix.h"
#include "ieee_fp.h"
#include "garmin_id.h"

enum {
	DLE           =  16,
	ETX           =   3.
	PID_ACK       =   6,
	PID_NAK       =  21,
	PID_HELLO     = 254,  /* "product request" */
	PID_PRODUCT   = 255,
	PID_PROTOCOLS = 253,
};

static word
	PID_records,
	PID_complete,
	PID_position,
	PID_time,
	PID_wpt,
	PID_pvt,
	PID_command,

	/* PID_async_mask = 28, ??? */

	CMD_xfer_pos,
	CMD_xfer_time,
	CMD_xfer_wpt,
	CMD_start_pvt,

	TYPE_position,
	TYPE_time,
	TYPE_wpt,
	TYPE_pvt;

static void
L001_Link_Protocol_1()
{
	PID_records   = 27;
	PID_complete  = 12;
	PID_position  = 17;
	PID_time      = 14;
	PID_wpt       = 35;
	PID_pvt       = 51;
	PID_command   = 10;
}

static void
L002_Link_Protocol_2()
{
	PID_records   = 35;
	PID_complete  = 12;
	PID_position  = 24;
	PID_time      = 20;
	PID_wpt       = 43;
	PID_pvt       =     0;
	PID_command   = 11;
}

static void
A010_Command_Protocol_1()
{
	CMD_xfer_pos  =  2;
	CMD_xfer_time =  5;
	CMD_xfer_wpt  =  7;
	CMD_start_pvt = 49;
}

static void
A011_Command_Protocol_2()
{
	CMD_xfer_pos  =     0;
	CMD_xfer_time = 20;
	CMD_xfer_wpt  = 21;
	CMD_start_pvt =     0;
}

static void
garmin_protocols_default(word id, word version)
{
	L001_Link_Protocol_1();
	A010_Command_Protocol_1();
	TYPE_wpt = 100;

	TYPE_pvt =      0;

	switch (id) {

	case 77:
		TYPE_wpt = (version < 301) ? 100 : 103;
		break;

	case 29:
		TYPE_wpt = (version < 400) ? 101 : 102;
		break;

	case 44:
		TYPE_wpt = 101;
		break;

	case 72:
		TYPE_wpt = 104;
		break;

	case 48:
		TYPE_wpt = 154;
		break;

	case 71:
		TYPE_wpt = 155;
		break;

	case 39:
	case 15:
		TYPE_wpt = 151;
		break;

	case 49:
	case 76:
	case 88:
		TYPE_wpt = 102;
		break;

	case 36:
	case 22:
	case 45:
	case 50:
	case 53:
	case 112:
		TYPE_wpt = 152;
		break;

	case 73:
	case 87:
	case 95:
	case 96:
	case 97:
	case 100:
	case 105:
	case 106:
		TYPE_wpt = 103;
		break;

	case 20:
	case 33:
	case 34:
	case 52:
	case 64:
	case 98:
		L002_Link_Protocol_2();
		A011_Command_Protocol_2();
		TYPE_wpt = 150;
		break;
	}
}

static bool
garmin_pvt_fix_is_good(word fix)
{
#if 0
		return (fix > 2);   /* 3 means 2D */
	else
#endif
	return (fix >= 2); /* 2 means 2D */
}

struct D600_time {
	byte month, day;
	word year;
	byte hour, filler;
	byte minute, second;
};

struct D700_position {
	ieee_double
		lat, lon;           /* radians */
};

struct D800_pvt {
	ieee_float
		altitude,           /* meters above WGS-84 ellipsoid */
		epe, eph, epv;
	word
		fix;                /* 2 or more is good - unless an old gps */
	ieee_double
		tow,                /* seconds into week */
		lat, lon;           /* radians */
	ieee_float
		speed_east, speed_north, speed_up, /* meters / second */
		msl_height;         /* from MSL to ellipsoid */
	word
		leap_seconds;       /* seconds from GPS to UTC */
	dword
		wn_days;            /* GPS week as days from UTC 31.12.1989 */
};


static word
	garmin_product, garmin_version;

static byte
	garmin_cksum, fill2;


struct {
	struct vector v;
	long altitude;
	struct wallclock wc;
} garmin;

/*
 *  pi shifted by 29 is 0x6487ed51, so
 *  32.32 multiplier to get milliminutes:
 */
static long scaled_radian_to_mm = 27501974;

static void
rad_coord(ieee_double rad, long *pmm)
{
	*pmm = ieee_double_to_long(rad, 29);

	SCALE32_signed(pmm, &scaled_radian_to_mm);
}

static void
pvt_altitude(struct D800_pvt *pvt, long *palt)
{
	*palt = ieee_float_to_long(pvt->altitude, 0)
	      + ieee_float_to_long(pvt->msl_height, 0);
}

static void
pvt_course_speed(struct D800_pvt *pvt, struct vector *v)
{
	/* shorts are wide enough */

	short a = ieee_float_to_long(pvt->speed_east, 1); /* shifted to ~ knots */
	short b = ieee_float_to_long(pvt->speed_west, 1);

	v.course = iatan2(b, a);
	v.knots  = ihypot(abs(a), abs(b));
}

static void
pvt_clock(struct D800_pvt *pvt, struct wallclock *wc)
{
	long tow = ieee_double_to_long(pvt->tow, 0)
	         - ld_word_intel(&pvt->leap_seconds);

	/* shift epoch to Sun 2005-07-10 00:00:00 --- gps week 1331 */

	word day = ld_dword_intel(&pvt->wn_days) - 1331 * 7;

	word date[3];
	word hour, min, sec;

	if (tow < 0) {
		tow += 24L * 60 * 60;
		day--;
	}
	hour = EDIV(tow, 60 * 60, &sec);
	min  = EDIV(sec, 60, &sec);

	day += EDIV(hour, 24, &hour);

	date_from_days_since_2005_07_10(day, date);

	wc->year  = date[0];
	wc->month = date[1];
	wc->day   = date[2];

	wc->hour   = hour;
	wc->minute = min;
	wc->second = sec;
}

static void
garmin_pvt(struct D800_pvt *pvt)
{
	rad_coord(pvt->lat, &garmin.v.lat);
	rad_coord(pvt->lon, &garmin.v.lon);

	pvt_cse_spd(pvt, &garmin.v);

	pvt_altitude(pvt, &garmin.altitude);
	pvt_clock(pvt, &garmin.wc);
}

static void
garmin_time(struct D600_time *t)
{
	garmin.wc.month  = t->month;
	garmin.wc.day    = t->day;
	garmin.wc.year   = ld_word_intel(&t->year) % 100;
	garmin.wc.hour   = t->hour;
	garmin.wc.minute = t->minute;
	garmin.wc.second = t->second;
}

static void
garmin_position(struct D700_position *p)
{
	rad_coord(p->lat, &garmin.v.lat);
	rad_coord(p->lon, &garmin.v.lon);
}


/*---------------------------------------------------*/

#define ENCODE(fn, arg) static byte * encode_ ## fn (byte *b, arg)

ENCODE(zero, word cc)
{
	do {
		*b++ = 0;
	} while (--cc);
	return (b);
}

ENCODE(blank, word cc)
{
	do {
		*b++ = ' ';
	} while (--cc);
	return (b);
}

ENCODE(string, byte *s)
{
	while (*b++ = *s++)
		;
	return (b);
}

ENCODE(byte, byte x)
{
	*b = x;
	return (b + 1);
}

ENCODE(word, word x)
{
	*(word *) b = x;
	return (b + 2);
}

ENCODE(ident_6, byte *id) /* ident 6 blank padded characters */
{
	xxx 

	return (b + 6);
}

ENCODE(comment_40, byte *s)
{
	xxx 

	return (b + 40);
}

ENCODE(ident_string, byte *id)
{
	while (*b++ = *id++)
		;
	return (b);
}

ENCODE(semicircle, long *pmm) /* lat/lon     semicircle dword */
{
	xxx 

	return (b + 4);
}

ENCODE(symbol_byte_, byte s)
{
	*b = s;
	return (b + 1);
}

ENCODE(symbol_word_, word s)
{
	*(word *) b = s;
	return (b + 2);
}

ENCODE(dspl_0, byte d)
{
	*b = d;
	return (b + 1);
}

#undef  ENCODE
/*---------------------------------------------------*/
#define ENCODE(fn, arg) b = encode_ ## fn (b, arg)

static word
wpt_encode(byte *buf, struct waypoint *w)
{
	byte *b = buf;

	byte *id = w->tag;
	byte *comment = "";

	switch (TYPE_wpt) {

	case 100:
		ENCODE(ident_6, id);
		ENCODE(semicircle, &w->lat);
		ENCODE(semicircle, &w->lon);
		ENCODE(zero, 4);
		ENCODE(comment_40, comment);
		break;

	case 101:
		ENCODE(ident_6, id);
		ENCODE(semicircle, &w->lat);
		ENCODE(semicircle, &w->lon);
		ENCODE(zero, 4);
		ENCODE(comment_40, comment);
		ENCODE(zero, 4);
		ENCODE(symbol_byte_, 0);
		break;

	case 102:
		ENCODE(ident_6, id);
		ENCODE(semicircle, &w->lat);
		ENCODE(semicircle, &w->lon);
		ENCODE(zero, 4);
		ENCODE(comment_40, comment);
		ENCODE(zero, 4);
		ENCODE(symbol_word_, 0);
		break;

	case 103:
		ENCODE(ident_6, id);
		ENCODE(semicircle, &w->lat);
		ENCODE(semicircle, &w->lon);
		ENCODE(zero, 4);
		ENCODE(comment_40, comment);
		ENCODE(symbol_byte_, 0);
		ENCODE(dspl_0, 0);          /*  0 w/name 1 symbol only 2 w/comment */
		break;

	case 104:
		ENCODE(ident_6, id);
		ENCODE(semicircle, &w->lat);
		ENCODE(semicircle, &w->lon);
		ENCODE(zero, 4);
		ENCODE(comment_40, comment);
		ENCODE(zero, 4);
		ENCODE(symbol_word_, 0);
		ENCODE(dspl_0, 3);         /* 0/1 symbol only 3 w/name 5 w/comment */
		break;

	case 105:
		ENCODE(semicircle, &w->lat);
		ENCODE(semicircle, &w->lon);
		ENCODE(symbol_word_, 0);
		ENCODE(ident_string, id);
		break;

	case 106:
		ENCODE(byte, 0);             /* class   byte -- zero */
		ENCODE(zero, 13);            /* subclass */
		ENCODE(semicircle, &w->lat);
		ENCODE(semicircle, &w->lon);
		ENCODE(symbol_word_, 0);
		ENCODE(ident_string, id);
		ENCODE(string, "");
		break;

	case 107:
		ENCODE(ident_6, id);
		ENCODE(semicircle, &w->lat);
		ENCODE(semicircle, &w->lon);
		ENCODE(zero, 4);
		ENCODE(comment_40, comment);
		ENCODE(symbol_byte_, 0);
		ENCODE(dspl_0, 0);            /* see 103 */
		ENCODE(zero, 4);
		ENCODE(byte, 0);      /* color byte 0 default 1 red 2 green 3 blue */
		break;

	case 108:
		ENCODE(byte, 0);      /* class   byte -- zero */
		ENCODE(byte, 0);      /* color   byte 0 default 1 red 2 green 3 blue */
		ENCODE(dspl_0, 0);    /* see 103 */
		ENCODE(byte, 0);      /* attr    byte 0 default 1 red 2 green 3 blue */
		ENCODE(symbol_word_, 0);
		ENCODE(zero, 18);     /* subclass 18 characters -- uh, zero */
		ENCODE(semicircle, &w->lat);
		ENCODE(semicircle, &w->lon);
		ENCODE(zero, 4);      /* alt     float */
		ENCODE(zero, 4);      /* depth   float */
		ENCODE(zero, 4);
		ENCODE(zero, 2);      /* state   2 bytes */
		ENCODE(zero, 2);      /* cc      2 bytes country code */
		ENCODE(ident_string, id);
		ENCODE(string, comment);
		ENCODE(string, "");
		ENCODE(string, "");
		ENCODE(string, "");
		ENCODE(string, "");
		break;

	case 109:
		ENCODE(byte, 0x01);    /* dtyp    byte -- must be 0x01 */
		ENCODE(byte, 0);       /* class   byte -- zero */
		ENCODE(byte, 0);      /* color   byte 0 default 1 red 2 green 3 blue */
		ENCODE(byte, 0);      /* attr    byte 0 default 1 red 2 green 3 blue */
		ENCODE(symbol_word_, 0);
		ENCODE(zero, 18);      /* subclass 18 characters -- uh, zero */
		ENCODE(semicircle, &w->lat);
		ENCODE(semicircle, &w->lon);
		ENCODE(zero, 4);             /* alt     float */
		ENCODE(zero, 4);             /* depth   float */
		ENCODE(zero, 4);
		ENCODE(zero, 2);        /* state   2 bytes */
		ENCODE(zero, 2);        /* cc      2 bytes country code */
		ENCODE(zero, 4);
		ENCODE(ident_string, id);
		ENCODE(string, comment);
		ENCODE(string, "");
		ENCODE(string, "");
		ENCODE(string, "");
		ENCODE(string, "");
		break;

	case 110:
		ENCODE(byte, 0x01);   /* dtyp    byte -- must be 0x01 */
		ENCODE(byte, 0);      /* class   byte -- zero */
		ENCODE(byte, 0);      /* color   byte 0 default 1 red 2 green 3 blue */
		ENCODE(byte, 0);      /* attr    byte 0 default 1 red 2 green 3 blue */
		ENCODE(symbol_word_, 0);
		ENCODE(zero, 18);     /* subclass 18 characters -- uh, zero */
		ENCODE(semicircle, &w->lat);
		ENCODE(semicircle, &w->lon);
		ENCODE(zero, 4);             /* alt     float */
		ENCODE(zero, 4);             /* depth   float */
		ENCODE(zero, 4);
		ENCODE(zero, 2);             /* state   2 bytes */
		ENCODE(zero, 2);             /* cc      2 bytes country code */
		ENCODE(zero, 4);
		ENCODE(zero, 4);             /* temp    float temperature */
		ENCODE(zero, 4);             /* time    dword timestamp */
		ENCODE(word, 0);             /* wpt_cat word category */
		ENCODE(ident_string, id);
		ENCODE(string, comment);
		ENCODE(string, "");
		ENCODE(string, "");
		ENCODE(string, "");
		ENCODE(string, "");
		break;

	case 150:
		ENCODE(ident_6, id);
		ENCODE(zero, 2);             /* cc      2 bytes country code */
		ENCODE(byte, 0);             /* wpt_class byte */
		ENCODE(semicircle, &w->lat);
		ENCODE(semicircle, &w->lon);
		ENCODE(word, 0);             /* alt     word */
		ENCODE(blank, 24);           /* city    24 chars */
		ENCODE(zero, 2);             /* state   2 chars */
		ENCODE(blank, 30);           /* name    30 chars */
		ENCODE(comment_40, comment);
		break;

	case 151:
		ENCODE(ident_6, id);
		ENCODE(semicircle, &w->lat);
		ENCODE(semicircle, &w->lon);
		ENCODE(zero, 4);
		ENCODE(comment_40, comment);
		ENCODE(zero, 4);
		ENCODE(blank, 30);           /* name    30 chars */
		ENCODE(blank, 24);           /* city    24 chars */
		ENCODE(zero, 2);             /* state   2 chars */
		ENCODE(word, 0);             /* alt     word */
		ENCODE(zero, 2);             /* cc      2 bytes country code */
		ENCODE(byte, 0);             /* zero    byte */
		ENCODE(byte, 0);             /* wpt_class byte */
		break;

	case 152:
		ENCODE(ident_6, id);
		ENCODE(semicircle, &w->lat);
		ENCODE(semicircle, &w->lon);
		ENCODE(zero, 4);
		ENCODE(comment_40, comment);
		ENCODE(zero, 4);
		ENCODE(blank, 30);            /* name    30 chars */
		ENCODE(blank, 24);            /* city    24 chars */
		ENCODE(zero, 2);              /* state   2 chars */
		ENCODE(word, 0);              /* alt     word */
		ENCODE(zero, 2);              /* cc      2 bytes country code */
		ENCODE(byte, 0);              /* zero    byte */
		ENCODE(byte, 0);              /* wpt_class byte */
		break;

	case 154:
		ENCODE(ident_6, id);
		ENCODE(semicircle, &w->lat);
		ENCODE(semicircle, &w->lon);
		ENCODE(zero, 4);
		ENCODE(comment_40, comment);
		ENCODE(zero, 4);
		ENCODE(blank, 30);         /* name    30 chars */
		ENCODE(blank, 24);         /* city    24 chars */
		ENCODE(zero, 2);           /* state   2 chars */
		ENCODE(word, 0);           /* alt     word */
		ENCODE(zero, 2);           /* cc      2 bytes country code */
		ENCODE(byte, 0);           /* zero    byte */
		ENCODE(byte, 0);           /* wpt_class byte */
		ENCODE(symbol_word_, 0);
		break;

	case 155:
		ENCODE(ident_6, id);
		ENCODE(semicircle, &w->lat);
		ENCODE(semicircle, &w->lon);
		ENCODE(zero, 4);
		ENCODE(comment_40, comment);
		ENCODE(zero, 4);
		ENCODE(blank, 30);           /* name    30 chars */
		ENCODE(blank, 24);           /* city    24 chars */
		ENCODE(zero, 2);             /* state   2 chars */
		ENCODE(word, 0);             /* alt     word */
		ENCODE(zero, 2);             /* cc      2 bytes country code */
		ENCODE(byte, 0);             /* zero    byte */
		ENCODE(byte, 0);             /* wpt_class byte */
		ENCODE(symbol_word_, 0);
		ENCODE(dspl_0, 0);           /* see 103 (?) */
		break;
	}
	return (b - buf);
}

#undef ENCODE
/*---------------------------------------------------*/

static void
garmin_protocols(byte *p, word len)
{
	/*
	 *  struct { byte tag; word data; } protocols[]
	 */
	byte *end = &p[len] - 2; /* no careless step past end */
	word data, Annn, Dx;

	word protocols = 0;

	for ( ; p < end; p += 3) {

		data = ld_word_intel(p + 1);
		switch (*p) {

		/* Physical protocol Pdata --- ignore */
		case 'P':
			break;

		/* Link protocol Ldata */
		case 'L':
			switch (data) {
			case 1: protocols |= 1; break;
			case 2: protocols |= 2; break;
			}
			break;

		/* Application protocol Adata */
		case 'A':
			switch (data) {
			case 10: protocols |= 0x10; break;
			case 11: protocols |= 0x20; break;
			}
			Annn = data;
			Dx = 0;
			break;

		/* Data Type <Dx> for protocol Annn is Ddata */
		case 'D':
			if (Dx++ == 0)
				switch (Annn) {
				case 100: TYPE_wpt = data; break; /* A100 waypoint transfer */
				case 800: TYPE_pvt = data; break; /* A800 PVT */
				}
			break;
		}
	}
	/* prefer Link and Device Command Protocols 1 and 1 */

	if ((protocols & 3) != 2)
		L001_Link_Protocol_1();
	else
		L002_Link_Protocol_2();

	if ((protocols & 0x30) != 0x20)
		A010_Command_Protocol_1();
	else
		A011_Command_Protocol_2();
}

static void
garmin_putc(byte c)
{
	garmin_cksum -= c;          /* nice trick for negated cksum */

	if (c == DLE) {
		SCI_putc(DLE);
		SCI_putc(DLE);
	} else
		SCI_putc(c);
}

static void
garmin_putw(word x)
{
	garmin_putc(x);
	garmin_putc(x >> 8);
}

static void
garmin_putbuf(byte *p, word n)
{
	do {
		garmin_putc(*p++);
	} while (--n);
}

static void
garmin_prologue(byte id, byte data_size)
{
	garmin_cksum = 0;

	SCI_putc(DLE);
	garmin_putc(id);
	garmin_putc(data_size);
}

static void
garmin_epilogue()
{
	garmin_putc(garmin_cksum);   /* nice trick for negated cksum */
	SCI_putc(DLE);
	SCI_putc(ETX);
}

static void
garmin_simple(byte id, word data)
{
	garmin_prologue(id, 2);  /* 2 bytes in word data */
	garmin_putw(data);
	garmin_epilogue();
}

static void
garmin_send_waypoint(struct waypoint *wpt)
{
	byte buf[256];
	word sz;

	if (CMD_xfer_wpt == 0)
		return;             /* no can do */

	sz = wpt_encode(buf, wpt);
	if (sz == 0)
		return;             /* unknown encoding */

	garmin_simple(PID_records, 1);

	garmin_prologue(PID_wpt, sz);
	garmin_putbuf(buf, sz);
	garmin_epilogue();

	garmin_simple(PID_complete, CMD_xfer_wpt);
}

static void
garmin_incoming(byte *msg)
{
	byte id = msg[0];
	byte sz = msg[1];

	msg += 2;

	/* something undocumented */

	if (id == 0)
		return;

	/* expect lots of these. no acks needed */

	if (id == PID_pvt && sz >= sizeof(struct garmin_pvt)) {
		garmin_pvt(msg);
		return;
	}
	if (id != PID_ACK
	 && id != PID_NAK)
		garmin_simple(PID_ACK, id);

	if (id == PID_position) {
		garmin_position(msg);
		return;
	}
	if (id == PID_time) {
		garmin_time(msg);
		return;
	}
	if (id == PID_PRODUCT && sz >= 4) {
		garmin_product = ld_word_intel(msg + 0);
		garmin_version = ld_word_intel(msg + 2);

		garmin_protocols_default(garmin_product, garmin_version);
		return;
	}
	if (id == PID_PROTOCOLS) {
		garmin_protocols(msg, sz);
		return;
	}
}

void
garmin_gps_input(byte c)
{
	static byte nDLE, sum, buf[256]; /* [ id data_size data* cksum ] */
	static word i;

	if (c == DLE) {
		++nDLE;
		return;                      /* on hold */
	}
	while (nDLE >= 2) {              /* DLE DLE --- a single DLE */
		nDLE -= 2;
		sum += buf[i++ % sizeof(buf)] = DLE;
	}
	if (nDLE && c != ETX)
		nDLE = sum = i = 0;          /* DLE id  --- rewind for new message */

	if (nDLE) {                      /* DLE ETX --- a complete message */
		nDLE = 0;
		if (sum == 0 && i >= 3 && i - 3 == buf[1])
			garmin_incoming(buf);
	} else
		sum += buf[i++ % sizeof(buf)] = c;
}

void
garmin_gps_init()
{
	init_SCI(9600, 0);

	/*
	 *  response to HELLO is
	 *  PID_PRODUCT    product, version, description
	 *  PID_PROTOCOLS  protocols[]
	 *  ... catch them later.
	 */
	garmin_prologue(PID_HELLO, 0);
	garmin_epilogue();

	/* mean time, some sensible defaults */

	L001_Link_Protocol_1();
	A010_Command_Protocol_1();

	garmin_simple(PID_command, CMD_start_pvt);
}

